#define USE_ALUGRID USE_ALUGRID
#define USE_AMG_SSOR USE_AMG_SSOR

#ifdef HAVE_CONFIG_H
# include "config.h"     
#endif
#include<iostream>
#include<sstream>

#include"dune/common/mpihelper.hh" // An initializer of MPI
#include"dune/common/exceptions.hh" // We use exceptions
#include<dune/common/fvector.hh>

#include<dune/grid/yaspgrid.hh>
#if HAVE_ALUGRID
#include<dune/grid/alugrid.hh>
#include<dune/grid/io/file/dgfparser/dgfalu.hh>
#include<dune/grid/io/file/dgfparser/dgfparser.hh>
#include<dune/grid/alugrid/3d/alugrid.hh>
#include<dune/grid/alugrid/3d/alu3dgridfactory.hh>
#endif

#include<dune/istl/bvector.hh>
#include<dune/istl/operators.hh>

#include<dune/pdelab/common/function.hh>
#include<dune/pdelab/common/functionutilities.hh>
#include<dune/pdelab/common/vtkexport.hh>
#include<dune/pdelab/finiteelementmap/p0fem.hh>
#include<dune/pdelab/finiteelementmap/q1fem.hh>
#include<dune/pdelab/finiteelementmap/conformingconstraints.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspace.hh>
#include<dune/pdelab/gridfunctionspace/gridfunctionspaceutilities.hh>
#include<dune/pdelab/gridfunctionspace/genericdatahandle.hh>
#include<dune/pdelab/gridfunctionspace/interpolate.hh>
#include<dune/pdelab/gridfunctionspace/constraints.hh>
#include<dune/pdelab/gridoperator/gridoperator.hh>
#include<dune/pdelab/gridoperatorspace/gridoperatorspace.hh>
#include<dune/pdelab/backend/istlvectorbackend.hh>
#include<dune/pdelab/backend/istlmatrixbackend.hh>
#include<dune/pdelab/backend/istlsolverbackend.hh>
#include<dune/pdelab/backend/novlpistlsolverbackend.hh>
#include<dune/pdelab/stationary/linearproblem.hh>

#include"bctype.hh"
#include"localoperatorfem.hh"

/**** Parameters ****/

/* dimension of grid */
const int dim=3;

/* relative reduction of residual */
const double resreduction=1.0e-5;

/* maximal number of iterations */
const int maxiter = 100;

/* Vertical anisotropy */
const double epsilon2 = 1.0e4;

/* Number of smoothing steps */
const double smoothsteps = 1;

/* verbosity level */
const int verbose = 2;

int main(int argc, char** argv)
{
  try{
    //Maybe initialize Mpi
    Dune::MPIHelper& helper = Dune::MPIHelper::instance(argc, argv);
    std::cout << "Hello World! This is dune-alugridnovlpamg." << std::endl;
    if(Dune::MPIHelper::isFake)
      std::cout<< "This is a sequential program." << std::endl;
    else
      std::cout<<"I am rank "<<helper.rank()<<" of "<<helper.size()
        <<" processes!"<<std::endl;

    /**** Create grid ****/
    Dune::FieldVector<double,dim> Size(1.0);
    Dune::FieldVector<int,dim> N(16);
    Dune::FieldVector<bool,dim> periodic(false);
#ifdef USE_ALUGRID
    std::cout << "Using ALUGrid" << std::endl;
    std::stringstream Filename;
    Filename << "cubedsphere.alu";
    Filename << ".";
    Filename << helper.rank();
    typedef Dune::ALUCubeGrid<3,3> Grid;
    Grid grid(Filename.str(),helper.getCommunicator());
#else
    std::cout << "Using YaspGrid" << std::endl;
    typedef Dune::YaspGrid<dim> Grid;
    int overlap = 0;
    Grid grid(helper.getCommunicator(),Size,N,periodic,overlap);
#endif
    grid.globalRefine(1);
    typedef Grid::LeafGridView GV;
    GV gv = grid.leafView();
    

    typedef GV::Grid::ctype Coord;
    typedef double Real;
  
    typedef Dune::PDELab::LinearSolverResult<double> Result;
    Result result;

    /**** Create Grid function space ****/
    typedef Dune::PDELab::Q1LocalFiniteElementMap<Coord, Real, dim> FEM;
    FEM fem;
    typedef Dune::PDELab::NonoverlappingConformingDirichletConstraints CON;
    typedef Dune::PDELab::ISTLVectorBackend<1> VBE;
    typedef Dune::PDELab::GridFunctionSpace<GV, FEM, CON, VBE> GFS;

    CON con;
    GFS gfs(gv, fem, con);
    con.compute_ghosts(gfs);
    
    typedef BCType<GV> B;
    B b(gv);
    typedef GFS::ConstraintsContainer<Real>::Type CC;
    CC cc;
    Dune::PDELab::constraints(b,gfs,cc);

    /**** Create grid operator ****/
    typedef LocalOperatorFEM<B> LOP;
    LOP lop(b);
    typedef Dune::PDELab::ISTLBCRSMatrixBackend<1,1> MBE;
    typedef Dune::PDELab::GridOperator<GFS,GFS,LOP,MBE,Real,Real,Real,CC,CC,true> GO;
    GO go(gfs,cc,gfs,cc,lop);

    /**** Solution vector ****/
    typedef GO::Traits::Domain U;
    U u(gfs, 0.0);

    /**** Solve linear problem ****/
#ifdef USE_AMG_SSOR
    std::cout << "Using ISTLBackend_NOVLP_CG_AMG_SSOR" << std::endl;
    typedef Dune::PDELab::ISTLBackend_NOVLP_CG_AMG_SSOR<GO> LS;
    LS ls (gfs, maxiter, verbose);
#else
    std::cout << "Using ISTLBackend_NOVLP_CG_SSORk" << std::endl;
    typedef Dune::PDELab::ISTLBackend_NOVLP_CG_SSORk<GO> LS;
    LS ls (gfs, maxiter,smoothsteps,verbose);
#endif
    Dune::PDELab::StationaryLinearProblemSolver<GO,LS,U>
      slp(go,u,ls, resreduction);
    slp.apply();
    result = ls.result();
  
    /**** Write out solution ****/
    typedef Dune::PDELab::DiscreteGridFunction<GFS,U> DGF;
    DGF udgf(gfs,u);
    Dune::VTKWriter<GV> vtkwriter(gv, Dune::VTKOptions::conforming);
    vtkwriter.addVertexData(new Dune::PDELab::VTKGridFunctionAdapter<DGF>(udgf,"solution"));
    vtkwriter.pwrite("solution",".",".",Dune::VTKOptions::ascii);
  }
  catch (Dune::Exception &e){
    std::cerr << "Dune reported error: " << e << std::endl;
  }
  catch (...){
    std::cerr << "Unknown exception thrown!" << std::endl;
  }
}
