/* ************************************************************************** *
 *
 *
 * ************************************************************************** */
#ifndef LOCALOPERATORFEM_HH
#define LOCALOPERATORFEM_HH LOCALOPERATORFEM_HH

#include<math.h>
#include<dune/grid/common/genericreferenceelements.hh>
#include<dune/grid/common/quadraturerules.hh>
#include<dune/pdelab/common/geometrywrapper.hh>
#include<dune/pdelab/localoperator/defaultimp.hh>
#include<dune/pdelab/localoperator/pattern.hh>
#include<dune/pdelab/localoperator/flags.hh>

/* ************************************************************************** *
 * ************************************************************************** */

template<class B>
class LocalOperatorFEM : 
  public Dune::PDELab::NumericalJacobianApplyVolume<LocalOperatorFEM<B> >,
  public Dune::PDELab::NumericalJacobianVolume<LocalOperatorFEM<B> >,
  public Dune::PDELab::NumericalJacobianApplySkeleton<LocalOperatorFEM<B> >,
  public Dune::PDELab::NumericalJacobianSkeleton<LocalOperatorFEM<B> >,
  public Dune::PDELab::NumericalJacobianApplyBoundary<LocalOperatorFEM<B> >,
  public Dune::PDELab::NumericalJacobianBoundary<LocalOperatorFEM<B> >,
  public Dune::PDELab::FullSkeletonPattern,
  public Dune::PDELab::FullVolumePattern,
  public Dune::PDELab::LocalOperatorDefaultFlags 
{
  public:
    enum { doPatternVolume = true };
    enum { doAlphaVolume = true };
    enum { doAlphaBoundary = true };
        
/* ********************************************************************** *
 *   Constructor
 * ********************************************************************** */
    LocalOperatorFEM (const B& b_,
                      unsigned int intorder_ = 2)
    : b(b_),
      intorder(intorder_)
    {}
    
/* ********************************************************************** *
 *   Local volume integral
 * ********************************************************************** */
    template<typename EG, typename LFSU, typename X, typename LFSV, typename R>
    void alpha_volume(const EG& eg,            // Volume element
                      const LFSU& lfsu,        // Trial function space
                      const X& x,              // Local coefficient of the expansion in trial function space
                      const LFSV& lfsv,        // Test function space
                      R& r                     // Local contribution to residual
                     ) const {
      typedef typename LFSV::Traits::FiniteElementType::Traits::LocalBasisType::Traits::DomainFieldType DF;
      typedef typename LFSV::Traits::FiniteElementType::Traits::LocalBasisType::Traits::RangeFieldType RF;
      typedef typename LFSV::Traits::FiniteElementType::Traits::LocalBasisType::Traits::JacobianType JacobianType;
      typedef typename LFSV::Traits::FiniteElementType::Traits::LocalBasisType::Traits::RangeType RangeType;
      typedef typename LFSU::Traits::SizeType size_type;
      const int dim = EG::Geometry::dimension;
      const int dimw = EG::Geometry::dimensionworld;
      
      Dune::GeometryType gt = eg.geometry().type();
      const Dune::QuadratureRule<DF,dim>& rule = Dune::QuadratureRules<DF,dim>::rule(gt,intorder);
      for (typename Dune::QuadratureRule<DF,dim>::const_iterator it = rule.begin(); it !=rule.end(); ++it) {
        std::vector<RangeType> phi(lfsu.size());
        lfsu.finiteElement().localBasis().evaluateFunction(it->position(),phi);
        RF u = 0.0;
        for (size_type i=0; i<lfsu.size(); i++)
          u += x[i]*phi[i];
        std::vector<JacobianType> js(lfsu.size());
        lfsu.finiteElement().localBasis().evaluateJacobian(it->position(),js);
        const Dune::FieldMatrix<DF,dimw,dim>
          jac = eg.geometry().jacobianInverseTransposed(it->position());
        std::vector<Dune::FieldVector<RF,dim>> gradphi(lfsu.size());
        for (size_type i=0; i<lfsu.size(); i++)
          jac.mv(js[i][0],gradphi[i]);
        Dune::FieldVector<RF,dim> gradu(0.0);
        for (size_type i=0; i<lfsu.size();i++)
          gradu.axpy(x[i],gradphi[i]);
        Dune::FieldVector<RF,dim> globalpos = eg.geometry().global(it->position());
        RF a = 0.0;
        RF f = 1.0;
        RF factor = it->weight()*eg.geometry().integrationElement(it->position());
        for (size_type i=0; i<lfsu.size();i++)
          r[i] += (gradu*gradphi[i] + a*u*phi[i] - f*phi[i])*factor;
      }
    }
    
/* ********************************************************************** *
 *  Boundary integral
 * ********************************************************************** */
    template <typename IG, typename LFSU, typename X, typename LFSV, typename R>
    void alpha_boundary(const IG& ig,          // Face element
                        const LFSU& lfsu_s,    // Trial function space
                        const X& x_s,          // Function expansion coefficients
                        const LFSV& lfsv_s,    // Test function space
                        R& r_s) const {        // Residual (OUT)
      // Do nothing as for Neumann BCs the flux is zero and for Dirichlet BCs
      // the field is zero.
    }
      
/* ********************************************************************** *
 *  class fields
 * ********************************************************************** */
    const B& b;
    unsigned int intorder;                   // integration order
};
#endif
